#include "korg.hpp"
#include "korgtext.hpp"
#include <string.h>

char* pcgname;
FILE* pcgf;
PCGHEAD pcghead;
PCGDATA pcgdata;

const char* getprog(int i)
{
char *progplace = "Dxx";

  if (i < 64)
  {
    progplace[1] = (i / 8)+'1';
    progplace[2] = (i % 8)+'1';
  }
  else
  {
    progplace[1] = 'r';
    progplace[2] = (i % 8)+'7';
  }
  return (const char*)progplace;
}

void dumpprognames()
{
int count, i;
KORGNAME s;

  count = (int)(pcghead.prognames.len / KORGNAME_SIZE);

  printf("%d PROGRAM NAMES: \n", count);

  fseek(pcgf, pcghead.prognames.adr, SEEK_SET);
  for (i = 0; i < count && i < 64; i++)
  {
    if (!read_korgname(pcgf, s))
      return;
    printf("  %4s  %10s", getprog(i), s);
    if (i % 4 == 3)
      printf("\n");
  }
  printf("\n");
}

void pagehdr(int i, char* s)
{
  printf("\nPROG %10.10s %s %s\n", pcgdata.name, getprog(i), s);
}

void dumpdrumkits()
{
int kit, drum, count, i;
PCGDRUM d;

  count = (int)(pcghead.drumkits.len / sizeof(PCGDRUM));

  printf("%d DRUM KITS of %d drums: \n", count / DRUMSPERKIT, DRUMSPERKIT);
  if (sizeof(PCGDRUM) != 22)
    putchar(7);

  if (pcghead.drumkits.len % DRUMSPERKIT != 0)
    putchar(7);

  fseek(pcgf, pcghead.drumkits.adr, SEEK_SET);

  for (i = 0; i < count; i++)
  {
    kit = i / DRUMSPERKIT;
    drum = i % DRUMSPERKIT;

    if (!KORGREAD(pcgf, d))
      return;

    if (drum == 0)
      printf("DRUMKIT Dr%d\n", kit+7);

    printf("#%02d  ", drum);
    if (d.drum == 0)
      printf("---");
    else
      printf("%03d", d.drum-1);

    printf("   %-3s  %-04d L%+03d D%+03d %+3s Pan=%-3s %s | ",
      note2str(d.key, -1),
      -1, d.level, d.decay,
      group2str(d.group),
      pan2str(d.panlevel),
      efflev2str(d.effect_sendlevel, 0)
      );

    dump(d.d1, sizeof(d.d1));
    dump(&d.d2, sizeof(d.d2));
    dump(d.d3, sizeof(d.d3));

    printf("\n");
  }
}

void dumpvdf(VDF & vdf, int baseoct)
{
  printf("Cutoff=%02d  EG Intensity=%02d  Color=%02d\n",
    vdf.cutoff, vdf.intensity, vdf.kbdcolor);
  printf("KBD Track: Amount=%+03d     Key=%-3s %s\n",
     vdf.kbd_trk_amount, note2str(vdf.pivot_key, baseoct), mode2str(vdf.mode_vdf));
  printf("           Time=%02d  AT:%c DT:%c ST:%c RT:%c\n",
     vdf.kbd_trk_time, pm2char(vdf.kbd_timeflags, 0), pm2char(vdf.kbd_timeflags, 1), pm2char(vdf.kbd_timeflags, 2), pm2char(vdf.kbd_timeflags, 3));
  printf("Vel.Sense: Amount=%+03d       Color=%+03d\n",
     vdf.vel_sense_amount, vdf.velcolor);
  printf("           Time=%02d  AT:%c DT:%c ST:%c RT:%c\n",
     vdf.vel_sense_time, pm2char(vdf.vel_timeflags, 0), pm2char(vdf.vel_timeflags, 1), pm2char(vdf.vel_timeflags, 2), pm2char(vdf.vel_timeflags, 3));
  printf("AT%02d A%+03d DT%02d B%+03d ST%02d S%+03d RT%02d R%+03d\n",
     vdf.attacktime, vdf.attacklevel, vdf.decay_time, vdf.break_point,
     vdf.slope_time, vdf.sustain_level, vdf.release_time,
     vdf.release_level);
}

void dumpvda(VDF& vdf, int baseoct)
{
  printf("KBD Track: Amount=%+03d     Key=%-3s %s\n",
     vdf.vda.kbd_trk_amount, note2str(vdf.vda.key, baseoct), mode2str(vdf.mode_vda, 1));
  printf("           Time=%02d  AT:%c DT:%c ST:%c RT:%c\n",
     vdf.vda.kbd_trk_time, pm2char(vdf.vda_kbd_timeflags, 0), pm2char(vdf.vda_kbd_timeflags, 1), pm2char(vdf.vda_kbd_timeflags, 2), pm2char(vdf.vda_kbd_timeflags, 3));
  printf("Vel.Sense: Amount=%+03d\n",
     vdf.vda.vel_sense_amount);
  printf("           Time=%02d  AT:%c DT:%c ST:%c RT:%c\n",
     vdf.vda.vel_sense_time, pm2char(vdf.vda_vel_timeflags, 0), pm2char(vdf.vda_vel_timeflags, 1), pm2char(vdf.vda_vel_timeflags, 2), pm2char(vdf.vda_vel_timeflags, 3));
  printf("AT%02d A%+03d DT%02d B%+03d ST%02d S%+03d RT%02d\n",
     vdf.vda.attacktime, vdf.vda.attacklevel, vdf.vda.decay_time, vdf.vda.break_point,
     vdf.vda.slope_time, vdf.vda.sustain_level, vdf.vda.release_time
     );
}

void dumpvdfmg()
{
  printf("VDFMG %s  Freq=%02d DL%02d I:%02d\n",
    waveform2str(pcgdata.vdfmg_waveform),
    pcgdata.vdfmg_freq,
    pcgdata.vdfmg_delay,
    pcgdata.vdfmg_intens);
  printf("      OSC:%s        AT%02d JS%02d S:%s\n",
    onoff2str(pcgdata.vdfmg_osc1),
    pcgdata.vdfmg_aftertouch,
    pcgdata.vdfmg_joystick,
    sync2str(pcgdata.vdfmg_sync));
  printf("\n");
  printf("Pitch Bend: Joystick=%+03d Aftertouch=%+03d\n",
    pcgdata.bend_joystick,
    pcgdata.bend_aftertouch);
  printf("    Cutoff: Joystick=%+03d Aftertouch=%+03d\n",
    pcgdata.cutoff_joystick,
    pcgdata.cutoff_aftertouch);
  printf("     Level: Aftertouch=%+03d\n",
    pcgdata.level_aftertouch);
}

void dumpeffect()
{
  printf("FX1 %02d:%-13s:%s Mod:%-5s I:%+03d\n",
    pcgdata.fx1, effect2str(pcgdata.fx1),
    onoff2str(pcgdata.fxflags & 3),
    dmod2str(pcgdata.fx1_dmod_source),
    pcgdata.fx1_dmod_amount);
  printf("each effect has own params\n");
  printf("\n");
  printf("FX2 %02d:%-13s:%s Mod:%-5s I:%+03d\n",
    pcgdata.fx2, effect2str(pcgdata.fx2),
    onoff2str(pcgdata.fxflags & 12),
    dmod2str(pcgdata.fx2_dmod_source),
    pcgdata.fx2_dmod_amount);
  printf("each effect has own params\n");
  printf("[ %9s ] C Pan = %02d:%02d  D Pan = %02d:%02d\n",
    "?????????", 0, 0, 0, 0);
}

void dumpprogdata()
{
int count, i, drum;

  count = (int)(pcghead.progdata.len / sizeof(PCGDATA));

  printf("%d PROGRAM DATA: \n", count);
  if (sizeof(PCGDATA) != 164)
    putchar(7);

  fseek(pcgf, pcghead.progdata.adr, SEEK_SET);

  for (i = 0; i < count; i++)
  {
    if (!KORGREAD(pcgf, pcgdata))
      return;

    drum = i >= 64;
    pagehdr(i, "1: OSCILLATOR");
    printf("%s %s  Hold %s  %sphonic\n",
      drum ? "Drum" : (pcgdata.doubleoscillator ? "Double" : "Single"),
      drum ? "" : "Oscillator",
       pcgdata.hold ? "On " : "Off",
      pcgdata.monophonic ? "Mono" : "Poly");
    printf("%03d: %10s L%02d %2d' PEG%+03d Pan=%s\n",
      pcgdata.osc1_multisound, multisound2str(pcgdata.osc1_multisound),
      pcgdata.vdf1.level, 0,
       pcgdata.vdf1.pitch.intens, pan2str(pcgdata.vdf1.panpot));
    printf("                              %s\n",
      efflev2str(pcgdata.vdf1.effect_send_level_c, pcgdata.vdf1.effect_send_level_d));

    if (!drum && pcgdata.doubleoscillator)
    {
      printf("%03d: %10s L%02d %2d' PEG%+03d Pan=%s\n",
	pcgdata.osc2_multisound, multisound2str(pcgdata.osc2_multisound),
	pcgdata.vdf2.level, 0,
	 pcgdata.vdf2.pitch.intens, pan2str(pcgdata.vdf2.panpot));
      printf("Intvl=%+03d Detune=%+03d Delay=%02d C=%d  D=%d\n",
	pcgdata.intervall, pcgdata.detune, pcgdata.delay,
	pcgdata.vdf2.effect_send_level_c,
	pcgdata.vdf2.effect_send_level_d);
    }
    printf("S%+03d AT%02d A%+03d DT%02d RT%02d R%+03d L%+03d T%+03d\n",
      pcgdata.startlevel, pcgdata.attacktime, pcgdata.attacklevel,
      pcgdata.decaytime, pcgdata.releasetime, pcgdata.releaselevel,
      pcgdata.velsense, pcgdata.timevelsense
      );
    pagehdr(i, "2: VDF 1");
    dumpvdf(pcgdata.vdf1, -2);

    if (!drum && pcgdata.doubleoscillator)
    {
      pagehdr(i, "3: VDF 2");
      dumpvdf(pcgdata.vdf2, 1);
    }
    pagehdr(i, "4: VDA 1");

    dumpvda(pcgdata.vdf1, -1);
    if (!drum && pcgdata.doubleoscillator)
    {
      pagehdr(i, "5: VDA 2");
      dumpvda(pcgdata.vdf2, 1);
    }

    pagehdr(i, "7: VDF MG/JS");
    dumpvdfmg();

    pagehdr(i, "8: EFFECT");
    dumpeffect();
  }
}

void dumppcg()
{

  if (!pcgheader(pcgf, pcghead))
  {
    fprintf(stderr, "%s is not a KORG pcg file\n", pcgname);
    return;
  }
  printf("\nKORG PCG file %s\n\n", pcgname);

  dumpprognames();

  dumpdrumkits();

  dumpprogdata();
}

void main(int argc, char**argv)
{
  argc--; argv++;

  if (argc == 0)
    return;

  pcgname = *argv++; argc--;

  pcgf = fopen(pcgname, "rb");
  if (!pcgf)
  {
    perror(pcgname);
    return;
  }

  dumppcg();

  fclose(pcgf);
}
